
// Miscellaneous global operations.

#include <stdlib.h>
#include <stdio.h>
extern "C" {
  #include <stdarg.h>
}
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "Procs.h"
#include "ProcTree.h"
#include "threads.h"
#include "schedule.h"
#include "TreeSchedule.h"
#include "adtags.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Global `Adlib' operations---initialisation and finalisation.        //
//                                                                     //
// This module also provides wrappers for a couple of low-level        //
// message-passing operations, used occasionally outside the           //
// communication-handling layer proper (the `ProcTree' and `Schedule'  //
// classes).                                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


Procs* world ;
//ActiveProcessGroup apg ;
TreeSchedule AdlibSpan ;


// Following two operations are globally collective

void AdlibInit(int& argc, char** (&argv)) {
  MPI_Init(&argc, &argv) ;

  world = new Procs ;
  //apg = new ActiveProcessGroup(*world) ;
  //apg.init(world) ;
  apg = *world ;
  apgStack.init() ;

  //apg.spanTree(&AdlibSpan) ;
  apgSpanTree(&AdlibSpan) ;

  Server :: threadsInit() ;

  RangeInit() ;
}

void AdlibFinalize() {

  Server :: threadsFinalize() ;

  delete world ;

  MPI_Finalize() ;

  exit(0) ;  // shouldn't be necessary, but...
}
 
void AdlibPrintf(char* message, ...) {

  // **** This won't work for floating point numbers!! ****

  va_list ap ;

  {
    int locId ;
    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;

    printf("Process %d: ", locId) ;
  }

  va_start(ap, message) ;

  char* substring = message ;
  int datum = 0 ;  // Initial value ignored, but keeps compiler quiet.
  for(char* ch = message ; ; ch++)
    switch(*ch) {
      case 0 :
        printf(substring, datum) ;
        return ;
      case '%' :
        {
          int sublen = ch - substring ;
          char* subcopy = new char [sublen + 1] ;

          memcpy(subcopy, substring, sublen) ;
          subcopy [sublen] = 0 ;
          printf(subcopy, datum) ;

          delete [] subcopy ;
        }
        substring = ch ;
        datum = va_arg(ap, int) ;  // No double-word arguments!
        break ;
      default :
        break ;
    }

  va_end(ap) ;
}

void AdlibError(char* message) {
    printf("%s\n", message) ;
  MPI_Abort(MPI_COMM_WORLD, 1) ;
}


int AdlibNProcs() {
  int size ;
  MPI_Comm_size(MPI_COMM_WORLD, &size) ;

  return size ;
}

int AdlibLocId() {
  int coord ;
  MPI_Comm_rank(MPI_COMM_WORLD, &coord) ;

  return coord ;
}

void AdlibSend(const int blockSize, char* data, const int dstId) {
  MPI_Status status ;

  MPI_Isend(data, blockSize, MPI_CHAR, dstId, DAT, MPI_COMM_WORLD, event) ;
  Server :: waitClientEvent(&status) ;
}

int AdlibRecv(const int blockSize, char* data, const int srcId) {
  MPI_Status status ;

  MPI_Irecv(data, blockSize, MPI_CHAR, srcId, DAT, MPI_COMM_WORLD, event) ;
  Server :: waitClientEvent(&status) ;

  int count ;
  MPI_Get_count(&status, MPI_CHAR, &count) ;

  return count ;
}

void AdlibBroadcast(const int blockSize, char* data) {
  //AdlibSpan->ProcTree :: broadcast(blockSize, data) ;
  AdlibSpan.broadcast(blockSize, data) ;
}

// -------------------------------------------------------------------
//
// Things to do:
//
//   Improve the implementation of `AdlibPrintf', as per `gprintf'.
//       (Rename it to `lprintf', and move it into `io.C'?)

